#include <cassert>
#include <cstdlib>
#include <dlfcn.h>
#include <iostream>
#include <sstream>
#include "NimGame.hpp"
#include "NimMove.hpp"
#include "NimPlayer.hpp"
#include "UniformIntDistribution.hpp"

using NimPlayerConstructor = NimPlayer*();

NimPlayer* load_player(char* name) {
   std::string path("./"); // look in current directory
   path += name;
   path += ".so"; // add suffix
   void* handle = dlopen(path.c_str(), RTLD_LAZY | RTLD_LOCAL);
   if (!handle) return nullptr;
   NimPlayerConstructor* constructor =
      (NimPlayerConstructor*) dlsym(handle, "construct");
   if (!constructor) {
      dlclose(handle); return nullptr;
   }
   return constructor();
}

unsigned int string_to_value(const char* cp) {
   std::string s(cp);
   std::istringstream is(s);
   unsigned int value;
   if (!(is >> value)) {
      std::cerr << "Unsigned integer value expected but got "
	 << cp << std::endl;
      std::exit(1);
   }
   return value;
}

int main(int argc, char** argv) {
   const char* cmdname = *argv++; --argc;
   if (argc != 4) {
      std::cerr << "Usage: " << cmdname <<
	 " #heaps maxsticks player#1 player#2" << std::endl;
      std::exit(1);
   }
   unsigned int number_of_heaps = string_to_value(*argv++); --argc;
   if (number_of_heaps == 0) {
      std::cerr << "Number of heaps must be positive" << std::endl;
      std::exit(1);
   }
   unsigned int maxtake = string_to_value(*argv++); --argc;
   NimPlayer* player[2] = {load_player(argv[0]), load_player(argv[1])};
   if (!player[0] || !player[1]) {
      std::cerr << "Dynamic loading of players failed. Bye." << std::endl;
      std::exit(1);
   }

   std::cout << "*** Game of Nim ***" << std::endl;

   // read parameters of the game
   std::cout << "Number of heaps: " << number_of_heaps << std::endl;
   std::cout << "Maximal number of sticks that can be taken in one move: "
      << maxtake << std::endl;

   // setup pseudo random generator
   UniformIntDistribution rand;

   // setup game
   NimGame game(number_of_heaps, maxtake);
   unsigned int minsize; unsigned int maxsize;
   if (maxtake) {
      minsize = maxtake * 2; maxsize = maxtake * 4;
   } else {
      minsize = 1; maxsize = 7;
   }
   unsigned int range = maxsize - minsize + 1;
   for (unsigned int i = 0; i < number_of_heaps; ++i) {
      game.set_heap_size(i, rand.draw(range) + minsize);
   }

   while (!game.finished()) {
      // print current state
      std::cout << "Heaps:";
      for (unsigned int i = 0; i < number_of_heaps; ++i) {
	 std::cout << " " << game.get_heap_size(i);
      }
      std::cout << std::endl;
      NimMove move;
      NimPlayer* next_player = player[game.get_next_player()];
      move = next_player->get_move(game);
      std::cout << next_player->get_name();
      if (move.has_resigned()) {
	 std::cout << " resigns." << std::endl;
      } else {
	 std::cout << " takes " << move.get_count() << " from heap "
	 << move.get_heap() << std::endl;
      }
      game.execute_move(move);
   }
   NimPlayer* winner = player[game.winner()];
   std::cout << winner->get_name() << " has won!" << std::endl;

   // make valgrind happy
   delete player[0]; delete player[1];
}
